#!/usr/bin/python
# -*- coding: utf-8 -*-

import appier

from .. import parts

class RipePulsePart(parts.Part):

    def login(self):
        self.id.login()

    def login_wait(self, redirect_url = None):
        self.id.login_wait(redirect_url = redirect_url)

    def logout(self):
        self.interactions.goto_url(self.signout_url, redirect_url = self.home_url)

    def logout_wait(self, redirect_url = None):
        redirect_url = redirect_url or self.home_url
        self.logout()
        self.waits.redirected_to(redirect_url)

    def click_order(self, number):
        self.interactions.click(".table .id a[href^='/orders/%d']" % number)
        self.waits.redirected_to(self.order_url(number))
        self.waits.visible(".title", text = "Order #%d" % number)

    def order_url(self, number):
        url = "%s/orders/%d" % (self.pulse_url, number)
        return url

    @property
    def pulse_url(self):
        ripe_suffix = appier.conf("RIPE_SUFFIX", None)
        if ripe_suffix: pulse_url = "https://ripe-pulse-%s.platforme.com" % ripe_suffix
        else: pulse_url = "http://localhost:3000"
        pulse_url = appier.conf("BASE_URL", pulse_url)
        pulse_url = appier.conf("PULSE_URL", pulse_url)
        pulse_url = appier.conf("RIPE_PULSE_URL", pulse_url)
        return pulse_url

    @property
    def base_url(self):
        return self.pulse_url

    @property
    def home_url(self):
        return "%s/" % self.pulse_url

    @property
    def next_url(self):
        return self.orders_url

    @property
    def orders_url(self):
        return "%s/orders" % self.pulse_url

    @property
    def signout_url(self):
        return "%s/signout" % self.pulse_url



#!/usr/bin/python
# -*- coding: utf-8 -*-

import time

import appier

import logging

from . import errors

class TestCase(appier.Observable):

    def __init__(self, *args, **kwargs):
        appier.Observable.__init__(self, *args, **kwargs)
        self.runner = kwargs.get("runner", None)
        self.loader = kwargs.get("loader", None)
        self.logger = kwargs.get("logger", self.__class__._build_logger(default = True))
        self.breadcrumbs = kwargs.get(
            "breadcrumbs",
            self.__class__._build_logger(
                name = "ripe-rainbow-breadcrumbs"
            )
        )
        self.browser_logger = kwargs.get(
            "browser_logger",
            self.__class__._build_logger(
                name = "ripe-rainbow-browser"
            )
        )

    @classmethod
    def _build_logger(
        cls,
        name = "ripe-rainbow",
        level = "INFO",
        silent = False,
        default = False
    ):
        # builds the unique instance base level name for the logger
        # and then verifies it is already registered in the class, if
        # that's the case returns the logger instance immediately
        name_i = "_logger_" + name
        memory_i = "_memory_" + name
        if hasattr(TestCase, name_i): return getattr(TestCase, name_i)

        # "computes" the proper verbosity level to be applied to the
        # standard output stream logger handler
        level = appier.conf("LEVEL", level)
        level = appier.conf("RAINBOW_LEVEL", level)
        level = logging.getLevelName(level.upper())

        # "calculates" the final value for the silent mode (no standard
        # stream handlers) taking the configuration values into consideration
        silent = appier.conf("SILENT", silent, cast = bool)
        silent = appier.conf("RAINBOW_SILENT", silent, cast = bool)

        formatter = logging.Formatter("%%(asctime)s [%s] [%%(levelname)s] %%(message)s" % name)

        logger = logging.getLogger(name)
        logger.parent = None
        logger.setLevel(logging.DEBUG)

        if not silent:
            handler = logging.StreamHandler()
            handler.setLevel(level)
            handler.setFormatter(formatter)
            handler._name = name
            logger.addHandler(handler)

        memory_handler = appier.MemoryHandler()
        memory_handler.setLevel(logging.DEBUG)
        memory_handler.setFormatter(formatter)
        memory_handler._name = name
        logger.addHandler(memory_handler)

        setattr(TestCase, name_i, logger)
        setattr(TestCase, memory_i, memory_handler)

        memory_handlers = getattr(TestCase, "_memory_handlers", [])
        memory_handlers.append(memory_handler)
        setattr(TestCase, "_memory_handlers", memory_handlers)

        return logger

    def before(self):
        pass

    def after(self):
        memory_handlers = getattr(TestCase, "_memory_handlers", [])
        for memory_handler in memory_handlers:
            memory_handler.clear()

    def succeeded(self, test, ctx = None):
        pass

    def failed(self, test, exception, ctx = None):
        pass

    def skipped(self, test):
        pass

    def run_all(self):
        for test in self.tests():
            self.run_test(test)

    def run_test(self, test, ctx = None):
        self.before()
        try:
            test()
            self.succeeded(test, ctx = ctx)
        except errors.SkipError:
            self.skipped(test)
            raise
        except Exception as exception:
            self.failed(test, exception, ctx = ctx)
            raise
        finally:
            self.after()

    def error(self, message):
        if not self.logger: return
        self.logger.error(message)

    def warning(self, message):
        if not self.logger: return
        self.logger.warning(message)

    def info(self, message):
        if not self.logger: return
        self.logger.info(message)

    def debug(self, message):
        if not self.logger: return
        self.logger.debug(message)

    def log_stack(self, method = None):
        pass

    def pause(self, timeout = 86400):
        time.sleep(timeout)

    def skip(self, message = None, reason = None):
        raise errors.SkipError(
            message = message,
            reason = reason or message
        )

    @property
    def tests(self):
        return [
            getattr(self, name) for name in dir(self)
            if not name == "tests" and hasattr(getattr(self, name), "test")
        ]



#!/usr/bin/python
# -*- coding: utf-8 -*-

import appier

from .. import parts

class LogicPart(parts.Part):

    def at_url(
        self,
        url,
        params = None,
        fragment = None,
        starts_with = False,
        strict = False
    ):
        # retrieves the current URL as a string from the underlying
        # driver so that it can be verified
        current_url = self.driver.current_url

        # runs the URL verification taking into account the URL currently
        # set in the driver (the interactive browser)
        return self.match_url(
            current_url,
            url,
            params = params,
            fragment = fragment,
            starts_with = starts_with,
            strict = strict
        )

    def match_url(
        self,
        url,
        expected,
        params = None,
        fragment = None,
        starts_with = False,
        strict = False
    ):
        # in case the URL is not defined, it only matches if it is
        # expected to be empty or not defined, this is a special
        # situation and early exit is used to avoid the raising of
        # an exception latter on (would pose issues in workflow)
        if url == None: return expected == None

        # runs the normalization process for the provided parameters,
        # so that they can be readily compared with the parsed ones
        params = self._normalize_params(params)

        # parses the current URL in the browser and reconstructs it with just
        # the base scheme and the URL path, makes sure that the provided URL
        # value is a valid and consistent string (avoids mixing bytes with strings)
        url = appier.legacy.str(url)
        url_p = appier.legacy.urlparse(url)
        url_params = appier.http._params(url_p.query)
        url_base = url_p.scheme + "://" + url_p.netloc + url_p.path

        # in case the provided URL is not a sequence converts it into
        # one so that it can be used in the underlying algorithm, then
        # makes sure that all of its components are properly encoded
        # as valid string values (avoiding byte related issues)
        if not isinstance(expected, (list, tuple)): expected = (expected,)
        expected = tuple([appier.legacy.str(value) for value in expected])

        # iterates over the complete set of (expected) URLs to be tested and
        # sees if at least one of them validates against the provided base URL
        for _url in expected:
            # in case the starts with mode is active verifies that the current
            # URL starts with the current expected in iteration
            if starts_with and not url.startswith(_url): continue

            # otherwise tries to run the regex match operation (by asserting
            # that the match method is present)
            elif hasattr(_url, "match") and not _url.match(url): continue

            # otherwise runs the "default" net location and path based verification
            # so that the "initial" part of the URL is validated (no query of fragment)
            else:
                _url_p = appier.legacy.urlparse(_url)
                _url_base = _url_p.scheme + "://" + _url_p.netloc + _url_p.path
                if not _url_base == url_base: continue

            # runs the extra set of verification (parameters and fragment) in
            # case they have been requested (proper parameters set)
            if not params == None and not\
                self._compare_params(url_params, params, strict = strict): continue
            if not fragment == None and not url_p.fragment == fragment: continue

            # returns a valid value as the current URL in iteration complies
            # with the complete set of items for acceptance criteria
            return True

        # builds the string representation of the expected value(s) so that it's
        # easier to diagnose possible problems in the URL expectation
        expected_s = " OR ".join(expected)
        args_l = []
        if params: args_l.append("params = %s" % params)
        if fragment: args_l.append("fragment = %s" % fragment)
        args_l.append("starts_with = %s" % str(starts_with))
        args_l.append("strict = %s" % str(strict))
        args_s = ", ".join(args_l)
        expected_s += " (%s)" % args_s

        # prints a debug message on the breadcrumbs logger indicating that there's
        # a missmatch in the verification of the URL
        self.breadcrumbs.debug("Provided URL is '%s' and not '%s'" % (url, expected_s))

        # returns the invalid value as none of the expected URL was able to be validated
        # against the provided URL
        return False

    def has_text(self, element, selector, text, safe = True):
        text = appier.legacy.u(text)

        if not element: return None

        # tries to retrieve the text (value) from the element taking into consideration
        # the kind of element that is being validation
        if element.tag_name == "input":
            # retrieves the DOM value attribute for the element as the
            # textual representation for it
            element_text = element.get_attribute("value")
        else:
            # scroll the current viewport to the element selection so that it's
            # safer to verify the text content of it, the underlying driver under
            # some conditions requires viewport visibility to verify the text
            if safe: self.driver.safe(self.driver.scroll_to, element)

            # retrieves the visual textual representation of the element as the
            # value for the element text, this value is only guaranteed to be valid
            # in case the element is visible on the viewport (display not none and
            # opacity greater than zero)
            element_text = element.text

        if not element_text == text:
            self.breadcrumbs.debug("Element '%s' found but has text '%s' instead of '%s'" % (
                selector,
                element_text,
                text
            ))

            return None

        self.breadcrumbs.debug("Found element '%s' with text '%s'" % (
            selector,
            text
        ))

        return element

    def has_src(self, element, url, params = None):
        src = element.get_attribute("src")
        return self.logic.match_url(src, url, params = params)

    def get(self, selector, condition = None):
        # tries to retrieve the complete set of elements that match
        # the provided selector and fulfill the condition
        matching = self.find(selector, condition = condition)

        # verifies if there are too many elements selected and raises
        # a warning as that probably indicated that the selector is broad
        if len(matching) > 1:
            self.breadcrumbs.warning(
                "Found more than one element for selector '%s', "
                "more specific selector is required" % selector
            )

        # returns the properly selected element in case there are valid
        # matches or an invalid value otherwise
        return matching[0] if len(matching) > 0 else None

    def find(self, selector, condition = None):
        # determines if there's a valid condition provided and if that's
        # not the case sets the default condition value
        has_condition = True if condition else False
        if not condition: condition = lambda e, s: True

        # runs the selection operation using the underlying driver
        elements = self.driver.safe(self.driver.find_elements, selector)

        # in case no elements match the provided selector then returns the
        # empty sequence immediately
        if not elements or len(elements) == 0:
            self.breadcrumbs.debug("Could not find elements with '%s'" % selector)
            return []

        # runs the filtering operation so that only the elements that match
        # the provided condition are selected (requires at least one to pass)
        elements = [element for element in elements if condition(element, selector)]

        # iterates over the complete set of selected elements to monkey patch
        # them with the extra domain specific symbols to allow extra information
        # and interaction to be accessible at an element level
        for element in elements:
            self._patch_element(
                element,
                selector = selector,
                condition = condition
            )

        # prints the proper debug message for diagnostics taking into account
        # the kind of selection that has been performed in the elements
        if len(elements) == 0:
            self.breadcrumbs.debug("Found elements with '%s' but none matches the condition" % selector)
        elif has_condition:
            self.breadcrumbs.debug("Found elements with '%s' that match the condition" % selector)
        else:
            self.breadcrumbs.debug("Found elements with '%s'" % selector)

        # returns the complete set of element that exist in the current context
        # and that match the requested condition
        return elements

    def _normalize_params(self, params):
        if not params: return params
        return dict((key, value if isinstance(value, (list, tuple)) else [value]) for\
            key, value in appier.legacy.iteritems(params))

    def _compare_params(self, first, second, strict = False):
        if strict: return first == second
        for key in second:
            if not key in first: return False
            is_sequence = isinstance(first[key], (list, tuple)) and\
                isinstance(second[key], (list, tuple))
            is_equal = sorted(first[key]) == sorted(second[key]) if\
                is_sequence else first[key] == second[key]
            if not is_equal: return False
        return True

    def _patch_element(self, element, selector = None, condition = None):
        def _is_same():
            _element = self.get(element._selector, condition = element._condition)
            return _element.id == element.id

        def _ensure_same(message = None):
            if element._is_same(): return
            message = message or "Element (%s) is not the same" % element._text()
            raise appier.OperationalError(message = message)

        def _text():
            description_s = " %s" % element._condition_description if element._condition_description else ""
            return "%s%s" % (element._selector, description_s)

        def _attr(key, value):
            if value: return self.driver.instance.execute_script("arguments[0].%s = \"%s\";" % (key, value), element)
            return self.driver.instance.execute_script("return arguments[0].%s;" % key, element)

        def _highlight():
            self.driver.highlight(element)

        def _lowlight():
            self.driver.lowlight(element)

        def _scroll_to():
            self.driver.scroll_to(element)

        element._selector = selector
        element._condition = condition
        element._condition_description = getattr(condition, "_description", None)

        element._is_same = _is_same
        element._ensure_same = _ensure_same
        element._text = _text
        element._attr = _attr
        element._highlight = _highlight
        element._lowlight = _lowlight
        element._scroll_to = _scroll_to
        element._focus = _scroll_to



#!/usr/bin/python
# -*- coding: utf-8 -*-

import appier

from .. import parts

class RipeUtilVuePart(parts.Part):

    def login_wait(self, environment = "ci", redirect_url = None):
        self._login_wait(redirect_url)
        self.waits.visible(".top-bar .center .selector .title")
        self.interactions.goto_url(
            redirect_url or self.home_url,
            params = [
                ("env", [environment])
            ]
        )
        self.waits.visible(".top-bar .center .selector .selection", text = environment)

    def logout_wait(self):
        self.interactions.goto_url(self.signout_url, redirect_url = (self.home_url, self.id.login_url))

    def _login_wait(self, redirect_url = None):
        self.interactions.goto_url(self.home_url, redirect_url = (self.home_url, self.id.login_url))
        self.id.authorize(redirect_url = self.next_url)

        redirect_url = redirect_url or self.home_url
        if self.interactions.url.startswith(redirect_url): return

        self.interactions.click(".form .button-blue")
        self.waits.redirected_to(redirect_url)

    @property
    def util_vue_url(self):
        util_vue_url = "https://ripe-util-vue.platforme.com"
        util_vue_url = appier.conf("BASE_URL", util_vue_url)
        util_vue_url = appier.conf("UTIL_VUE_URL", util_vue_url)
        util_vue_url = appier.conf("RIPE_UTIL_VUE_URL", util_vue_url)
        return util_vue_url

    @property
    def base_url(self):
        return self.util_vue_url

    @property
    def home_url(self):
        return "%s/" % self.util_vue_url

    @property
    def next_url(self):
        return self.home_url

    @property
    def signout_url(self):
        return "%s/signout" % self.util_vue_url

    @property
    def delivery_url(self):
        return "%s/delivery" % self.base_url

    @property
    def example_url(self):
        return "%s/example" % self.base_url

    @property
    def locales_url(self):
        return "%s/locales" % self.base_url

    @property
    def order_url(self):
        return "%s/order" % self.base_url

    @property
    def orders_url(self):
        return "%s/orders" % self.base_url

    @property
    def platforme_id_url(self):
        return "%s/platforme-id" % self.base_url

    @property
    def quote_url(self):
        return "%s/quote" % self.base_url

    @property
    def rules_url(self):
        return "%s/rules" % self.base_url

    @property
    def background_url(self):
        return "%s/background" % self.base_url



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import sys
import time

import appier

from .. import info
from .. import errors

LOG_LEVELS = (
    "FINEST",
    "FINER",
    "FINE",
    "INFO",
    "WARNING",
    "SEVERE"
)

LOG_LEVELS_M = dict(
    FINEST = "debug",
    FINER = "debug",
    FINE = "info",
    WARNING = "warn",
    SEVERE = "error"
)

class InteractiveDriver(object):

    def __init__(self, owner, **kwargs):
        self.owner = owner

    @staticmethod
    def driver_g(name):
        return globals()[name.capitalize() + "Driver"]

    @staticmethod
    def label_g():
        driver = appier.conf("DRIVER", "selenium")
        return InteractiveDriver.driver_g(driver).label()

    @classmethod
    def label(cls):
        return "Interactive"

    def start(self):
        pass

    def stop(self):
        pass

    def touch_diag(self, name, delta = 1):
        if not hasattr(self.owner, "diag"): return
        if self.owner.diag == None: return
        self.owner.diag[name] = self.owner.diag.get(name, 0) + delta

    def switch_context(self, name, *args, **kwargs):
        raise appier.NotImplementedError()

    def count_context(self, context, *args, **kwargs):
        raise appier.NotImplementedError()

    def clear_storage(self):
        self.clear_cookies()
        self.clear_local_storage()
        self.clear_session_storage()

    def clear_cookies(self):
        raise appier.NotImplementedError()

    def clear_local_storage(self):
        raise appier.NotImplementedError()

    def clear_session_storage(self):
        raise appier.NotImplementedError()

    def get_key(self, name):
        raise appier.NotImplementedError()

    def get(self, url):
        raise appier.NotImplementedError()

    def find_element(self, selector):
        raise appier.NotImplementedError()

    def find_elements(self, selector):
        raise appier.NotImplementedError()

    def find_by_name(self, name):
        raise appier.NotImplementedError()

    def press_key(self, element, key, ensure = True):
        raise appier.NotImplementedError()

    def write_text(self, element, text, ensure = True):
        raise appier.NotImplementedError()

    def click(self, element, ensure = True):
        raise appier.NotImplementedError()

    def ensure_visible(self, element, timeout = None):
        raise appier.NotImplementedError()

    def scroll_to(self, element, sleep = None):
        """
        Scrolls the element on which it's called into the visible area
        of the browser window.

        This operation should be performed in a deterministic way.

        :type element: Element
        :param element: The element to scroll into view.
        :type sleep: int
        :param sleep: The number of seconds to sleep after the scroll
        operation, important for smooth scrolls.
        """

        raise appier.NotImplementedError()

    def highlight(self, elements):
        raise appier.NotImplementedError()

    def lowlight(self, elements):
        raise appier.NotImplementedError()

    def screenshot(self, file_path):
        raise appier.NotImplementedError()

    def switch_tab(self, tab):
        raise appier.NotImplementedError()

    def close_tab(self, tab = None):
        raise appier.NotImplementedError()

    def wrap_outer(self, method, *args, **kwargs):
        return method(*args, **kwargs)

    def wrap_inner(self, method, *args, **kwargs):
        return method(*args, **kwargs)

    def safe(self, method, *args, **kwargs):
        return self.wrap_outer(self.wrap_inner, method, *args, **kwargs)

    def press_enter(self, element, ensure = True):
        return self.press_key(element, "enter", ensure = ensure)

    @property
    def options(self):
        raise appier.NotImplementedError()

    @property
    def current_url(self):
        raise appier.NotImplementedError()

    @property
    def current_tab(self):
        raise appier.NotImplementedError()

    @property
    def tab_count(self):
        raise appier.NotImplementedError()

    def _wait(self, timeout = None):
        raise appier.NotImplementedError()

class SeleniumDriver(InteractiveDriver):

    def __init__(self, owner, **kwargs):
        InteractiveDriver.__init__(self, owner, **kwargs)
        self.secure = appier.conf("SEL_SECURE", False, cast = bool)
        self.browser = appier.conf("SEL_BROWSER", "chrome")
        self.browser_cache = appier.conf("SEL_BROWSER_CACHE", True, cast = bool)
        self.maximized = appier.conf("SEL_MAXIMIZED", False, cast = bool)
        self.headless = appier.conf("SEL_HEADLESS", False, cast = bool)
        self.device = appier.conf("SEL_DEVICE", None)
        self.window_size = appier.conf("SEL_WINDOW_SIZE", "1920x1080")
        self.pixel_ratio = appier.conf("SEL_PIXEL_RATIO", 1, cast = int)
        self.mobile_emulation = appier.conf("SEL_MOBILE_EMULATION", False, cast = bool)
        self.poll_frequency = appier.conf("SEL_POLL_FREQUENCY", None, cast = float)
        self.service_args = appier.conf("SEL_SERVICE_ARGS", [], cast = list)
        self.secure = kwargs.get("secure", self.secure)
        self.browser = kwargs.get("browser", self.browser)
        self.browser_cache = kwargs.get("browser_cache", self.browser_cache)
        self.maximized = kwargs.get("maximized", self.maximized)
        self.headless = kwargs.get("headless", self.headless)
        self.device = kwargs.get("device", self.device)
        self.window_size = kwargs.get("resolution", self.window_size)
        self.window_size = kwargs.get("window_size", self.window_size)
        self.pixel_ratio = kwargs.get("pixel_ratio", self.pixel_ratio)
        self.mobile_emulation = kwargs.get("mobile", self.mobile_emulation)
        self.mobile_emulation = kwargs.get("mobile_emulation", self.mobile_emulation)
        self.poll_frequency = kwargs.get("poll_frequency", self.poll_frequency)
        self.service_args = kwargs.get("service_args", self.service_args)

    @classmethod
    def label(cls):
        import selenium.webdriver
        browser = cls.browser()
        return "Selenium %s (%s/%s)" % (
            selenium.webdriver.__version__,
            browser["name"],
            browser["version"]
        )

    @classmethod
    def browser(cls):
        import selenium.webdriver
        browser = appier.conf("SEL_BROWSER", "chrome")
        browser_cache = appier.conf("SEL_BROWSER_CACHE", True, cast = bool)
        fix_path = appier.conf("SEL_FIX_PATH", True, cast = bool)
        if fix_path: cls._fix_path()
        if browser == "chrome":
            options = selenium.webdriver.ChromeOptions()
            options.add_argument("--disable-gpu")
            options.add_argument("--disable-infobars")
            options.add_argument("--disable-extensions")
            options.add_argument("--disable-browser-side-navigation")
            options.add_argument("--headless")
            if not browser_cache:
                options.add_argument("--disable-application-cache")
            if sys.platform.startswith("linux"):
                options.add_argument("--disable-dev-shm-usage")
                options.add_argument("--no-sandbox")
            instance = selenium.webdriver.Chrome(options = options)
            try:
                name = instance.capabilities.get("browserName", None)
                version = instance.capabilities.get("browserVersion", None)
            finally:
                instance.quit()
        elif browser == "firefox":
            options = selenium.webdriver.FirefoxOptions()
            options.headless = True
            options.set_preference("devtools.console.stdout.content", True)
            instance = selenium.webdriver.Firefox(options = options)
            try:
                name = instance.capabilities.get("browserName", None)
                version = instance.capabilities.get("browserVersion", None)
            finally:
                instance.quit()
        else:
            raise appier.OperationalError(
                message = "Unknown browser '%s'" % browser
            )
        return dict(
            name = name,
            version = version
        )

    @classmethod
    def _fix_path(cls, paths = ("/usr/local/bin",)):
        path = os.environ.get("PATH", "")
        separator = ";" if os.name == "nt" else ":"
        path_s = path.split(separator)
        for _path in paths:
            if _path in path_s: continue
            path_s.append(_path)
        os.environ["PATH"] = separator.join(path_s)

    def stop(self):
        self._gc_tabs()
        self._flush_log()
        InteractiveDriver.stop(self)

    def clear_cookies(self):
        self.instance.delete_all_cookies()

    def clear_local_storage(self):
        self.instance.execute_script("localStorage.clear();")

    def clear_session_storage(self):
        self.instance.execute_script("sessionStorage.clear();")

    def get(self, url):
        # touches the diag values increment the number of operations
        # performed for the current execution context
        self.touch_diag("get")

        # performs the get URL operation redirecting the underlying
        # engine to the target web page
        return self.instance.get(url)

    def find_element(self, selector):
        return self.instance.find_element_by_css_selector(selector)

    def find_elements(self, selector):
        return self.instance.find_elements_by_css_selector(selector)

    def find_by_name(self, name):
        return self.instance.find_element_by_name(name)

    def find_element_by_css_selector(self, selector):
        return self.find_element(selector)

    def find_elements_by_css_selector(self, selector):
        return self.find_elements(selector)

    def find_element_by_name(self, name):
        return self.find_by_name(name)

    def press_key(self, element, key, ensure = True):
        # touches the diag values increment the number of operations
        # performed for the current execution context
        self.touch_diag("press_key")

        # in case the ensure flag is set makes sure that the element
        # is visible in an interactable way
        if ensure: self.ensure_visible(element)

        key = self._resolve_key(key)
        element.send_keys(key)
        return element

    def write_text(self, element, text, ensure = True):
        # in case the ensure flag is set makes sure that the element
        # is visible in an interactable way
        if ensure: self.ensure_visible(element)

        # sends the complete set of keys defined in the text
        # to the element and then returns it
        element.send_keys(text)
        return element

    def click(self, element, ensure = True):
        from selenium.common.exceptions import ElementClickInterceptedException, ElementNotVisibleException, WebDriverException

        # touches the diag values increment the number of operations
        # performed for the current execution context
        self.touch_diag("click")

        try:
            # in case the ensure flag is set makes sure that the element
            # is visible in an interactable way
            if ensure: self.ensure_visible(element)

            # moves the mouse to the element according to the pre-defined
            # pivot, this will allow proper clicking
            self._move_to(element)

            # runs the click operation directly on the element without any
            # kind of previous interaction as expected
            element.click()
        except (
            ElementClickInterceptedException,
            ElementNotVisibleException,
            WebDriverException
        ) as exception:
            self.owner.breadcrumbs.debug("Element is not \"clickable\" because: %s (%s)" % (
                exception,
                exception.__class__
            ))
            return None

        # returns the element object to the caller so that it can
        # be piped in a chain of operations
        return element

    def ensure_visible(self, element, timeout = None):
        # starts the operation by moving the cursor to the outside of the element
        # this ensures that the cursor is not "moving over" the element
        self._move_outside(element, raise_e = False)

        # "installs" the global event listeners for visibility detection
        # on the current target element, to be used by the try visible test
        self.__install_listener(element)

        try:
            self.wrap_outer(
                lambda: self._wait(timeout = timeout).until(
                    lambda d: self._try_visible(element),
                    "Element (%s) never became visible" % element._text()
                )
            )
        except errors.TimeoutError:
            try: self.__highlight(element)
            except Exception: pass
            raise
        finally:
            self.__uninstall_listener(element)

        # returns the element object so that in can be chained in multiple
        # operations (functional design decision)
        return element

    def scroll_to(self, element, position = "center", sleep = None):
        # builds the proper options string taking into account the offset
        # in terms of positioning of the element in the scroll
        if position == "center": options = "{ block: \"center\", inline: \"center\" }"
        else: options = ""

        # as Selenium doesn't automatically support automatically scrolling
        # in elements inside the page, such as when using an iframe or
        # overflow scroll, therefore we must rely on Web API `Element.scrollIntoView()`
        # that allows proper scroll operation into element
        self.instance.execute_script("arguments[0].scrollIntoView(%s);" % options, element)

        # when triggering a smooth scroll, the element may take some time to
        # be displayed in the desired position, hence the optional sleep
        if sleep: time.sleep(sleep)

        # returns the element back to the caller method, ready to be used
        # in a chained call strategy of execution
        return element

    def highlight(self, elements):
        if not isinstance(elements, (list, tuple)):
            elements = (elements,)
        for element in elements:
            self.__highlight(element)
        return elements

    def lowlight(self, elements):
        if not isinstance(elements, (list, tuple)):
            elements = (elements,)
        for element in elements:
            self.__lowlight(element)
        return elements

    def screenshot(self, file_path):
        self.instance.save_screenshot(file_path)

    def switch_tab(self, tab):
        try:
            new_window_handle = self.instance.window_handles[tab]
            self.instance.switch_to.window(new_window_handle)
        except IndexError:
            raise errors.UnexistingTabError(tab = tab, tab_count = self.tab_count)

    def close_tab(self, tab = None):
        tab = tab or self.current_tab

        if self.tab_count <= 1:
            raise errors.CloseTabError(
                message = "There is only a single tab, so you can't close it"
            )
        if tab >= self.tab_count:
            raise errors.UnexistingTabError(
                tab = tab,
                tab_count = self.tab_count
            )

        # saves the current tab in question so that we can restore
        # its selection at the end of the operation
        current_tab = self.current_tab

        tab_window_handle = self.instance.window_handles[tab]
        self.instance.switch_to.window(tab_window_handle)
        self.instance.close()

        if current_tab == tab: self.switch_tab(max(0, current_tab - 1))
        else: self.switch_tab(current_tab)

    def wrap_outer(self, method, *args, **kwargs):
        from selenium.common.exceptions import TimeoutException
        try:
            return method(*args, **kwargs)
        except TimeoutException as exception:
            raise errors.TimeoutError(message = exception.msg)

    def wrap_inner(self, method, *args, **kwargs):
        """
        Wraps the method being waited on to tolerate some exceptions since
        in most cases these are transitory conditions that shouldn't break
        the test.

        :type method: Function
        :param method: The method that is going to be executed in a wrapped
        fashion to properly handle exceptions.
        :rtype Function
        :return: The method wrapped on a try-catch for exceptions.
        """

        from selenium.common.exceptions import WebDriverException
        try: return method(*args, **kwargs)
        except (WebDriverException, AssertionError) as exception:
            self.owner.breadcrumbs.debug("Got exception while waiting: %s" % exception)
            return None

    @property
    def options(self):
        return dict(
            secure = self.secure,
            browser = self.browser,
            browser_cache = self.browser_cache,
            maximized = self.maximized,
            headless = self.headless,
            device = self.device,
            window_size = self.window_size,
            pixel_ratio = self.pixel_ratio,
            mobile_emulation = self.mobile_emulation,
            poll_frequency = self.poll_frequency,
            service_args = self.service_args
        )

    @property
    def current_url(self):
        return self.instance.current_url

    @property
    def current_tab(self):
        return self.instance.window_handles.index(self.instance.current_window_handle)

    @property
    def tab_count(self):
        return len(self.instance.window_handles)

    @property
    def instance(self):
        cls = self.__class__

        # in case there's already an instance defined in the class
        # and it is still considered valid then re-uses it
        if hasattr(cls, "_instance") and cls._instance and\
            hasattr(cls, "_options") and self.options == cls._options:
            return cls._instance

        import selenium.webdriver

        # in case there's an instance currently available destroys it
        # to avoid possible collision (garbage collection)
        self._destroy_instance()

        # "saves" the currently set options so that they can be used
        # in the definition of the instance to be created
        cls._options = self.options

        if self.browser == "chrome":
            # creates the underlying instance of the Chrome driver
            # that is going to be used in the concrete execution
            cls._instance = selenium.webdriver.Chrome(
                options = self._selenium_options(self.browser),
                desired_capabilities = self._selenium_capabilities(self.browser),
                service_args = self.service_args or None
            )
        elif self.browser == "firefox":
            # creates the underlying Firefox instance using the
            # pre-defined options as expected
            cls._instance = selenium.webdriver.Firefox(
                options = self._selenium_options(self.browser),
                desired_capabilities = self._selenium_capabilities(self.browser),
                firefox_profile = self._selenium_profile(self.browser),
                service_args = self.service_args or None
            )
        else:
            raise appier.OperationalError(
                message = "Unknown browser '%s'" % self.browser
            )

        # in case the browser should be started maximized, then
        # a new argument is added to the list of options
        if self.maximized:
            cls._instance.fullscreen_window()

        # in case the browser should have a specific window size
        # then splits the window size around the specific values
        if self.window_size:
            # unpacks the size into the integer components and runs
            # the set size operation at the driver level
            width, height = (int(value) for value in self.window_size.split("x"))
            cls._instance.set_window_size(width, height)

            # tries to obtain the new window size to make sure that
            # a proper resize operation has been performed
            size = cls._instance.get_window_size()
            appier.verify(
                size["width"] == width and size["height"] == height,
                message = "Invalid size, expected %s got %sx%s" %\
                    (self.window_size, size["width"] , size["height"])
            )

        # registers the destroy instance method to be called once
        # the runner is finished (proper cleanup)
        self.owner.runner.add_on_finish(self._destroy_instance)

        # returns the final instance of the driver to the caller
        # so that it can be used for operation
        return cls._instance

    def _destroy_instance(self):
        cls = self.__class__
        if not hasattr(cls, "_instance") or not self._instance:
            return
        cls._instance.quit()
        cls._instance = None
        cls._options = None

    def _selenium_options(self, browser = None):
        browser = browser or self.browser
        if not hasattr(self, "_selenium_options_%s" % browser): return None
        return getattr(self, "_selenium_options_%s" % browser)()

    def _selenium_options_chrome(self):
        import selenium.webdriver

        # crates the base object for the options to be used by
        # the Google Chrome browser
        options = selenium.webdriver.ChromeOptions()

        # in case the mobile emulation is required then an extra
        # experimental option is added to allow custom behaviour
        # (including touch, pixel ratio and user agent)
        if self.mobile_emulation:
            width, height = (int(value) for value in self.window_size.split("x", 1))
            user_agent = info.USER_AGENTS.get(self.device, None) if self.device else None
            emulation_options = dict(
                deviceMetrics = dict(
                    width = width,
                    height = height,
                    pixelRatio = self.pixel_ratio,
                    touch = False
                )
            )
            if user_agent: emulation_options["userAgent"] = user_agent
            options.add_experimental_option("mobileEmulation", emulation_options)

        # adds some of the default arguments to be used for the
        # execution of the Google Chrome instance
        options.add_argument("--disable-gpu")
        options.add_argument("--disable-infobars")
        options.add_argument("--disable-extensions")
        options.add_argument("--disable-browser-side-navigation")

        # in case the browser cache is disabled then add the correct
        # argument to the list of command line arguments
        if not self.browser_cache:
            options.add_argument("--disable-application-cache")

        # in case the headless mode was selected then an
        # extra argument is added to the set of options
        if self.headless:
            options.add_argument("--headless")

        # fixes issue when running within a docker container
        # since it typically uses /dev/shm shared memory with
        # little space making chrome crash when rendering large pages
        if sys.platform.startswith("linux"):
            options.add_argument("--disable-dev-shm-usage")
            options.add_argument("--no-sandbox")

        # returns the options to the calling method as expected
        # by the current infrastructure
        return options

    def _selenium_options_firefox(self):
        import selenium.webdriver

        # crates the base object for the options to be used by
        # the Mozilla Firefox browser
        options = selenium.webdriver.FirefoxOptions()

        # in case mobile emulation is requested the current test
        # must be skipped as there's no support available
        if self.mobile_emulation: self.owner.skip("Firefox can't run tests that require a mobile device")

        # in case the headless instance option is set propagates
        # it to the Firefox options object
        if self.headless:
            options.headless = True

        # returns the options to the calling method as expected
        # by the current infrastructure
        return options

    def _selenium_capabilities(self, browser):
        return getattr(self, "_selenium_capabilities_%s" % browser)()

    def _selenium_capabilities_chrome(self):
        from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

        capabilities = DesiredCapabilities.CHROME
        capabilities["loggingPrefs"] = dict(
            browser = "ALL",
            driver = "ALL",
            client = "ALL",
            performance = "ALL"
        )

        return capabilities

    def _selenium_capabilities_firefox(self):
        from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

        capabilities = DesiredCapabilities.FIREFOX
        capabilities["loggingPrefs"] = dict(
            browser = "ALL",
            driver = "ALL",
            client = "ALL",
            performance = "ALL"
        )

        return capabilities

    def _selenium_profile(self, browser):
        return getattr(self, "_selenium_profile_%s" % browser)()

    def _selenium_profile_firefox(self):
        from selenium.webdriver.firefox.firefox_profile import FirefoxProfile

        profile = FirefoxProfile()

        if not self.browser_cache:
            profile.set_preference("browser.cache.disk.enable", False)
            profile.set_preference("browser.cache.memory.enable", False)
            profile.set_preference("browser.cache.offline.enable", False)

        return profile

    def _wait(self, timeout = None):
        from selenium.webdriver.support.ui import WebDriverWait
        kwargs = dict()
        if self.poll_frequency: kwargs["poll_frequency"] = self.poll_frequency
        if timeout == None: timeout = self.owner.timeout
        return WebDriverWait(self.instance, timeout, **kwargs)

    def _move_to(self, element, pivot = "center"):
        element_width = element.size["width"]
        element_height = element.size["height"]

        if pivot == "center":
            return self._move_to_offset(
                element,
                x = int(element_width / 2),
                y = int(element_height / 2)
            )

        if pivot == "left-top":
            return self._move_to_offset(
                element,
                x = 0,
                y = 0
            )

        if pivot == "right-top":
            return self._move_to_offset(
                element,
                x = element_width - 1,
                y = 0
            )

        if pivot == "right-bottom":
            return self._move_to_offset(
                element,
                x = element_width - 1,
                y = element_height - 1
            )

        if pivot == "left-bottom":
            return self._move_to_offset(
                element,
                x = 0,
                y = element_height - 1
            )

    def _move_to_offset(self, element, x = 0, y = 0, safe = False):
        from selenium.webdriver.common.action_chains import ActionChains

        # creates an action chain that will move the mouse into the
        # target offset of coordinates relative to the element
        actions = ActionChains(self.instance)
        if safe: actions.move_to_element_with_offset(element, x, y)
        else: self._move_to_unsafe(actions, element, x, y)
        actions.perform()

        # returns the element that currently has the mouse pointing
        # to it according to the defined offset (if possible)
        return element

    def _move_to_unsafe(self, actions, element, x = 0, y = 0, duration = 5):
        instance = actions.w3c_actions.pointer_action

        if not x == None or not y == None:
            element_rect = element.rect
            left_offset = element_rect["width"] / 2
            top_offset = element_rect["height"] / 2
            left = left_offset * -1 + (x or 0)
            top = top_offset * -1 + (y or 0)
        else:
            left = 0
            top = 0

        instance.source.create_pointer_move(
            duration = duration,
            x = int(left),
            y = int(top),
            origin = element
        )

        return instance

    def _move_outside(self, element, raise_e = True):
        """
        Moves the mouse outside a given element, this should
        ensure proper outside of element position (not hover).

        This method is going to use a series of strategies to
        try to achieve the goal of moving the cursor outside
        of the referred element.

        :type element: Element
        :param element: The element to move outside from.
        :type raise_e: bool
        :param raise_e: If an exception should be raised when
        it's not possible to move outside the element.
        """

        from selenium.common.exceptions import MoveTargetOutOfBoundsException

        # gathers the element's dimensions to be able to calculate other
        # corner's offset positions, for the list of possibilities
        size = element.size
        width, height = size["width"], size["height"]

        # creates the tuple that contains the complete set of strategies
        # for the offset operation to move the cursor outside of the
        # requested element (avoiding collision)
        possibilities = (
            (-1, -1),
            (-1, 0),
            (0, -1),
            (width, -1),
            (width, 0),
            (-1, height),
            (0, height),
            (width, height)
        )

        # iterates over the complete set of offset possibilities
        # and runs the cursor to the requested offset, in case
        # there's no exception the control flow is returned (success)
        for x, y in possibilities:
            try:
                self._move_to_offset(element, x, y)
                return
            except MoveTargetOutOfBoundsException:
                pass

        # raises an operation error as it was not possible to move
        # the cursor outside of the current element using any of
        # the available strategies (only active if the raise~
        # exception flag is set as a parameter)
        if raise_e:
            raise appier.OperationalError(
                message = "Couldn't move outside element"
            )

    def _try_visible(self, element, strategy = "scroll_to"):
        from selenium.common.exceptions import MoveTargetOutOfBoundsException

        # makes sure that the element in testing is still the same that
        # has been originally selected (avoiding possible new elements with
        # the same selection criteria from being used)
        element._ensure_same()

        # prints some debug information on the retry of the visibility
        # test for the element in question
        self.owner.breadcrumbs.debug("Trying visibility on element '%s'" % element.id)

        try:
            # tries to run the cursor movement to the element, this may fail
            # as the element is not currently visible in the viewport
            self._move_to(element)
        except MoveTargetOutOfBoundsException:
            # some drivers raise exceptions when trying to move to elements
            # outside the viewport, we must ignore such exceptions
            pass

        # runs the is visible operation on the element that should guarantee
        # that the element is visible and interactable, if the returns is
        # positive the control flow is returned and the return value is positive
        if self.__is_visible(element): return True

        # executes the try-out strategy to try to make the element visible
        # in the view-port, by default this scrolls the element to the center
        # of the current screen (best possible strategy)
        if strategy == "scroll_to": self.scroll_to(element)

        # moves the element back to the outside of it, so that there's
        # a mouse movement one more time (skeptical move)
        self._move_outside(element, raise_e = False)

        # "resets" the value of the entered flag back to the original value
        # so that the test may be done against the movement to be performed
        self.instance.execute_script("window._entered = false")

        try:
            # tries to run the cursor movement to the element, this may or
            # may not fail according to the success of the strategy
            self._move_to(element)
        except MoveTargetOutOfBoundsException:
            # some drivers raise exceptions when trying to move to elements
            # outside the viewport, we must ignore such exceptions
            pass

        # verifies if the element is visible and interactable and if that's
        # the case returns a positive result and the control flow
        if self.__is_visible(element): return True

        # by default the visibility/interactability of the provided element
        # is assumed to be false (not interactable) as no tests have passed
        return False

    def _resolve_key(self, name):
        from selenium.webdriver.common.keys import Keys
        KEYS = dict(
            enter = Keys.ENTER,
            space = Keys.SPACE,
            backspace = Keys.BACKSPACE,
            left = Keys.LEFT,
            right = Keys.RIGHT,
            up = Keys.UP,
            down = Keys.DOWN,
            page_up = Keys.PAGE_UP,
            page_down = Keys.PAGE_DOWN,
            end = Keys.END,
            home = Keys.HOME,
            esc = Keys.ESCAPE,
            cmd = Keys.COMMAND,
            ctrl = Keys.CONTROL,
            tab = Keys.TAB,
            shift = Keys.SHIFT
        )
        return KEYS[name]

    def _flush_log(self, levels = LOG_LEVELS, browser = None):
        browser = browser or self.browser
        if not hasattr(self, "_flush_log_%s" % browser): return
        return getattr(self, "_flush_log_%s" % browser)()

    def _flush_log_chrome(self, levels = LOG_LEVELS):
        from selenium.common.exceptions import WebDriverException

        for name in ("browser", "client", "driver", "performance"):

            try: log = self.instance.get_log(name)
            except WebDriverException as exception:
                self.owner.browser_logger.warn(exception)
                log = []

            for item in log:
                if not item["level"] in levels: continue
                if not "message" in item: continue
                level, message = item["level"], item["message"]
                level_n = LOG_LEVELS_M.get(level, "info")
                level_m = getattr(self.owner.browser_logger, level_n)
                level_m(message.strip())

    def _gc_tabs(self):
        """
        Runs the garbage collection in the current window set, effectively
        closing all the extra windows currently visible.

        This is done by percolating over the old windows and closing them,
        and making sure that are the end the focus is at the main window.

        Running this operation ensures that the state of the browser is kept
        clean and ready to be used for new operations.
        """

        for window_handle in self.instance.window_handles[1:]:
            self.instance.switch_to.window(window_handle)
            self.instance.close()

        # in case there's no remaining window handles, then there's
        # nothing remaining to be done, returns the control flow
        if len(self.instance.window_handles) < 1: return

        # "grabs" the first (and main) window handle and switches
        # the control to it so that it can be properly used
        window_handle = self.instance.window_handles[0]
        self.instance.switch_to.window(window_handle)

    def __is_visible(self, element):
        """
        Internal method that verifies the visibility/interactability
        of the provided element taking into account a series of
        previously inserted setup operation.

        There are two strategies within this method one that uses the
        entered flag on window (previously installed handlers) and
        another (unsafer) that uses the :hover query selector.

        Should be used with proper care to avoid unwanted behaviour as
        it's considered an extremely low level function.

        :type element: Element
        :param element: The element to validated for visibility and/or
        intractability (probably for click purposes).
        :rtype: bool
        :return: If the element is currently interactable.
        """

        if not self.secure and self.instance.execute_script("return window._entered"): return True
        if self.instance.execute_script(
            "return Array.from(arguments[0].parentElement.querySelectorAll(\":hover\").values()).includes(arguments[0])",
            element
        ): return True
        return False

    def __install_listener(self, element):
        # sets the initial value of the "entered" global variable and then registers the mouse
        # over event listener that will change the entered flag value
        self.instance.execute_script("window._entered = false")
        self.instance.execute_script("window._handler = function() { window._entered = true; };")
        self.instance.execute_script("arguments[0].addEventListener(\"mouseenter\", window._handler, true);", element)
        self.instance.execute_script("arguments[0].addEventListener(\"mouseover\", window._handler, true);", element)

    def __uninstall_listener(self, element):
        # removes the complete set of global variables and event listeners associated with
        # the "global strategy" for visibility detection of an element
        self.instance.execute_script(
            "arguments[0].removeEventListener(\"mouseover\", window._handler);",
            element
        )
        self.instance.execute_script(
            "arguments[0].removeEventListener(\"mouseenter\", window._handler);",
            element
        )
        self.instance.execute_script("delete window._handler")
        self.instance.execute_script("delete window._entered")

    def __highlight(self, element, color = "#ff0000", sleep = None, safe = True):
        # in case the safe flag is set then both the transition and the
        # animation settings are disabled to make sure that the visual
        # updates are made immediately
        if safe:
            self.instance.execute_script(
                "arguments[0].style.transition = \"none\";",
                element,
            )
            self.instance.execute_script(
                "arguments[0].style.animation = \"none\";",
                element,
            )

        # changes the background color of the target element to make
        # it highlighted in contrast with the other elements
        self.instance.execute_script(
            "arguments[0].style.backgroundColor = \"%s\";" % color,
            element,
        )

        # in case the sleep value is set, then wait a certain amount of time
        # to make sure the UI operations are visible, taking into consideration
        # that due to things like transition and animations the visual operations
        # may take some time to be presented on screen
        if sleep: time.sleep(sleep)

    def __lowlight(self, element, sleep = None, safe = True):
        # in case the safe flag is set then both the transition and the
        # animation settings are re-enabled to make sure that the visual
        # updates are restored back to the original CSS values
        if safe:
            self.instance.execute_script(
                "arguments[0].style.transition = \"\";",
                element,
            )
            self.instance.execute_script(
                "arguments[0].style.animation = \"\";",
                element,
            )

        # removes the background color value back to the "original" unset value
        # so that top level CSS can take action
        self.instance.execute_script(
            "arguments[0].style.backgroundColor = \"\";",
            element,
        )

        # in case the sleep value is set, then wait a certain amount of time
        # to make sure the UI operations are visible, taking into consideration
        # that due to things like transition and animations the visual operations
        # may take some time to be presented on screen
        if sleep: time.sleep(sleep)

class AppiumDriver(InteractiveDriver):

    CONTEXT_MAPPER = dict(
        native = "NATIVE_APP",
        webview = "WEBVIEW_"
    )

    def __init__(self, owner, **kwargs):
        InteractiveDriver.__init__(self, owner, **kwargs)
        self.server_url = appier.conf("APM_SERVER_URK", "http://localhost:4723/wd/hub")
        self.avd = appier.conf("APM_AVD", "Nexus_5X_API_29_x86")
        self.platform = appier.conf("APM_PLATFORM", "Android")
        self.device = appier.conf("APM_DEVICE", "Android Emulator")
        self.package = appier.conf("APM_PACKAGE", "com.platforme.ripe_robin")
        self.activity = appier.conf("APM_ACTIVITY", "com.platforme.ripe_robin.MainActivity")
        self.app_path = appier.conf("APM_APP_PATH", "/Users/gcc/ripe-robin-revamp/android/app/build/outputs/apk/debug/app-debug.apk")
        self.headless = appier.conf("APM_HEADLESS", False, cast = bool)
        self.poll_frequency = appier.conf("APM_POLL_FREQUENCY", None, cast = float)
        self.server_url = kwargs.get("server_url", self.server_url)
        self.avd = kwargs.get("avd", self.avd)
        self.platform = kwargs.get("platform", self.platform)
        self.device = kwargs.get("device", self.device)
        self.package = kwargs.get("package", self.package)
        self.activity = kwargs.get("activity", self.activity)
        self.app_path = kwargs.get("app_path", self.app_path)
        self.headless = kwargs.get("headless", self.headless)
        self.poll_frequency = kwargs.get("poll_frequency", self.poll_frequency)

    @classmethod
    def label(cls):
        import appium.version
        return "Appium %s" % appium.version.version

    def switch_context(self, context, *args, **kwargs):
        if context == "native":
            return self.instance.switch_to.context(self.contexts[0])
        if context == "webview":
            index = kwargs.get("index", 0)
            webview_contexts = [context for context in self.contexts if context.startswith("WEBVIEW_")]
            return self.instance.switch_to.context(webview_contexts[index])
        raise appier.OperationalError(
            message = "Context not supported '%s'" % context
        )

    def count_context(self, context, *args, **kwargs):
        cls = self.__class__
        context_n = cls.CONTEXT_MAPPER.get(context, None)
        return len([context for context in self.contexts if context.startswith(context_n)])

    def find_element(self, selector):
        if self.in_native:
            element = None
            for method in (
                self.find_element_by_accessibility_id,
                self.find_element_by_id
            ):
                element = method(selector)
                if element: break
            return element
        else:
            return self.find_element_by_css_selector(selector)

    def find_elements(self, selector):
        if self.in_native:
            elements = []
            for method in (
                self.find_elements_by_accessibility_id,
                self.find_elements_by_id
            ):
                elements = method(selector)
                if len(elements) > 0: break
            return elements
        else:
            return self.find_elements_by_css_selector(selector)

    def find_element_by_css_selector(self, selector):
        return self.instance.find_element_by_css_selector(selector)

    def find_elements_by_css_selector(self, selector):
        return self.instance.find_elements_by_css_selector(selector)

    def find_element_by_accessibility_id(self, id):
        return self.instance.find_element_by_accessibility_id(id)

    def find_elements_by_accessibility_id(self, id):
        return self.instance.find_elements_by_accessibility_id(id)

    def find_element_by_id(self, id):
        return self.instance.find_element_by_id(id)

    def find_elements_by_id(self, id):
        return self.instance.find_elements_by_id(id)

    def press_key(self, element, key, ensure = True):
        # touches the diag values increment the number of operations
        # performed for the current execution context
        self.touch_diag("press_key")

        # in case the ensure flag is set makes sure that the element
        # is visible in an interactable way
        if ensure: self.ensure_visible(element)

        key = self._resolve_key(key)
        element.send_keys(key)
        return element

    def write_text(self, element, text, ensure = True):
        # in case the ensure flag is set makes sure that the element
        # is visible in an interactable way
        if ensure: self.ensure_visible(element)

        # sends the complete set of keys defined in the text
        # to the element and then returns it
        element.send_keys(text)
        return element

    def click(self, element, ensure = True):
        from selenium.common.exceptions import ElementClickInterceptedException, ElementNotVisibleException, WebDriverException

        # touches the diag values increment the number of operations
        # performed for the current execution context
        self.touch_diag("click")

        try:
            # in case the ensure flag is set makes sure that the element
            # is visible in an interactable way
            if ensure: self.ensure_visible(element)

            # moves the mouse to the element according to the pre-defined
            # pivot, this will allow proper clicking
            self._move_to(element)

            # runs the click operation directly on the element without any
            # kind of previous interaction as expected
            element.click()
        except (
            ElementClickInterceptedException,
            ElementNotVisibleException,
            WebDriverException
        ) as exception:
            self.owner.breadcrumbs.debug("Element is not \"clickable\" because: %s (%s)" % (
                exception,
                exception.__class__
            ))
            return None

        # returns the element object to the caller so that it can
        # be piped in a chain of operations
        return element

    def ensure_visible(self, element, timeout = None):
        return element

    def scroll_to(self, element, position = "center", sleep = None):
        return element

    def wrap_outer(self, method, *args, **kwargs):
        from selenium.common.exceptions import TimeoutException
        try:
            return method(*args, **kwargs)
        except TimeoutException as exception:
            raise errors.TimeoutError(message = exception.msg)

    def wrap_inner(self, method, *args, **kwargs):
        """
        Wraps the method being waited on to tolerate some exceptions since
        in most cases these are transitory conditions that shouldn't break
        the test.

        :type method: Function
        :param method: The method that is going to be executed in a wrapped
        fashion to properly handle exceptions.
        :rtype Function
        :return: The method wrapped on a try-catch for exceptions.
        """

        from selenium.common.exceptions import WebDriverException
        try: return method(*args, **kwargs)
        except (WebDriverException, AssertionError) as exception:
            self.owner.breadcrumbs.debug("Got exception while waiting: %s" % exception)
            return None

    @property
    def options(self):
        return dict(
            avd = self.avd,
            platform = self.platform,
            device = self.device,
            package = self.package,
            activity = self.activity,
            app_path = self.app_path,
            headless = self.headless,
            poll_frequency = self.poll_frequency
        )

    @property
    def instance(self):
        cls = self.__class__

        # in case there's already an instance defined in the class
        # and it is still considered valid then re-uses it
        if hasattr(cls, "_instance") and cls._instance and\
            hasattr(cls, "_options") and self.options == cls._options:
            return cls._instance

        import appium.webdriver

        # in case there's an instance currently available destroys it
        # to avoid possible collision (garbage collection)
        self._destroy_instance()

        # "saves" the currently set options so that they can be used
        # in the definition of the instance to be created
        cls._options = self.options

        # creates the underlying instance of the Appium driver
        # that is going to be used in the concrete execution
        cls._instance = appium.webdriver.Remote(
            self.server_url,
            desired_capabilities = dict(
                avd = self.avd,
                platformName = self.platform,
                deviceName = self.device,
                appPackage = self.package,
                appActivity = self.activity,
                app = self.app_path,
                isHeadless = self.headless
            )
        )

        # registers the destroy instance method to be called once
        # the runner is finished (proper cleanup)
        self.owner.runner.add_on_finish(self._destroy_instance)

        # returns the final instance of the driver to the caller
        # so that it can be used for operation
        return cls._instance

    @property
    def context(self):
        return self.instance.context

    @property
    def contexts(self):
        return self.instance.contexts

    @property
    def in_native(self):
        return self.context == "NATIVE_APP"

    @property
    def in_webview(self):
        return self.context.startswith("WEBVIEW_")

    def _destroy_instance(self):
        cls = self.__class__
        if not hasattr(cls, "_instance") or not self._instance:
            return
        cls._instance.quit()
        cls._instance = None
        cls._options = None

    def _wait(self, timeout = None):
        from selenium.webdriver.support.ui import WebDriverWait
        kwargs = dict()
        if self.poll_frequency: kwargs["poll_frequency"] = self.poll_frequency
        if timeout == None: timeout = self.owner.timeout
        return WebDriverWait(self.instance, timeout, **kwargs)

    def _move_to(self, element, pivot = "center"):
        return element

    def _flush_log(self, levels = LOG_LEVELS):
        from selenium.common.exceptions import WebDriverException

        for name in ("driver", "server"):
            try: log = self.instance.get_log(name)
            except WebDriverException as exception:
                self.owner.browser_logger.warn(exception)
                log = []

            for item in log:
                if not item["level"] in levels: continue
                if not "message" in item: continue
                level, message = item["level"], item["message"]
                level_n = LOG_LEVELS_M.get(level, "info")
                level_m = getattr(self.owner.browser_logger, level_n)
                level_m(message.strip())



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import sys
import inspect

import appier

from . import test_cases

class Loader(object):

    def test_suite(self, **kwargs):
        raise appier.NotImplementedError()

class PathLoader(Loader):

    def __init__(
        self,
        path = ".",
        extension = ".py",
        exclusion = ("setup.py",)
    ):
        self.path = path
        self.extension = extension
        self.exclusion = exclusion
        self.path = os.path.expanduser(self.path)
        self.path = os.path.abspath(self.path)
        self.path = os.path.normpath(self.path)

    def test_suite(self, **kwargs):
        return [test_cls(loader = self, **kwargs) for test_cls in self._load_classes(self.path)]

    def _load_classes(self, path, recursive = True):
        classes = []

        # runs the loading of the modules for the current path using
        # a possible recursive approach
        modules = self._load_modules(path, recursive = recursive)

        # iterates over the complete set of loaded modules to try
        # to load all of the test classes contained in them
        for module in modules:

            ctx = self._resolve_ctx(module)

            for name in dir(module):
                value = getattr(module, name)
                if not inspect.isclass(value): continue
                if not issubclass(value, test_cases.TestCase): continue
                value.ctx = ctx
                classes.append(value)

        return classes

    def _load_packages(self, path, recursive = True):
        packages = []

        if os.path.isdir(path):
            name = os.path.basename(path)
            dir_path = os.path.dirname(path)
            base_name = os.path.splitext(name)[0]
            sys.path.insert(0, dir_path)
            try:
                packages.append(__import__(base_name))
            except ImportError:
                pass
            finally:
                sys.path.remove(dir_path)

        names = os.listdir(path)
        for name in names:
            full_path = os.path.join(path, name)
            if os.path.isdir(full_path) and recursive:
                packages += self._load_packages(full_path, recursive = recursive)

        return packages

    def _load_modules(self, path, recursive = True):
        modules = []
        names = os.listdir(path)
        if not "." in sys.path: sys.path.insert(0, ".")
        for name in names:
            base_name = os.path.splitext(name)[0]
            full_path = os.path.join(path, name)
            if os.path.isdir(full_path) and recursive:
                modules += self._load_modules(full_path, recursive = recursive)
            elif not name in self.exclusion and name.endswith(self.extension):
                sys.path.insert(0, path)
                try:
                    modules.append(__import__(base_name))
                except (ImportError, ValueError):
                    pass
                finally:
                    sys.path.remove(path)
        return modules

    def _resolve_ctx(self, module):
        # uses the path to the module containing the test to try
        # to gather some context to the test, this means that the
        # name of the directory is going to be used as some kind
        # of context for the test execution
        module_path = module.__file__ if hasattr(module, "__file__") else None
        module_path = module_path or None
        if module_path:
            # creates the list that is going to hold the "chunks" that
            # are going to be joined to created the complete context
            ctx_l = []

            # starts the execution by "introducing" the name of the current
            # module's directory into the execution logic
            module_path = os.path.normpath(os.path.abspath(module_path))
            module_dir_path = os.path.dirname(module_path)
            module_dir_path = os.path.normpath(os.path.abspath(module_dir_path))

            # iterates over the complete set of parent package directories
            # (considering that a package is a directory that contains `__init__.py`)
            # until one that is not a parent one is found
            while True:
                if not module_dir_path: break
                if not "__init__.py" in os.listdir(module_dir_path): break
                ctx_l.insert(0, os.path.basename(module_dir_path))
                _module_dir_path = os.path.join(module_dir_path, "..")
                _module_dir_path = os.path.normpath(os.path.abspath(_module_dir_path))
                if _module_dir_path == module_dir_path: break
                module_dir_path = _module_dir_path

            # creates the context by joining the list of chunks present
            # in the context list
            ctx = ".".join(ctx_l)
        else:
            ctx = None
        return ctx



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import appier

from .. import parts

class InteractionsPart(parts.Part):

    def goto_url(self, url, redirect_url = None, params = [], fragment = "", wait = True):
        """
        Navigates to a certain URL with given GET parameters and to
        the request fragment.

        The operation is always going to be performed using a GET
        request as that's the only available request method from
        the browser.

        By default the redirection is waited to verify that the final
        redirection is performed.

        :type url: String
        :param url: The URL to navigate to, this should represent only
        the base URL without GET parameters and fragment.
        :type redirect_url: String
        :param redirect_url: The target URL of the redirection (if different
        from the target one) to be used in the wait verification.
        :type params: list
        :param params: A list of (key, values) tuples representing
        the GET query parameters to be added to the base URL
        :type fragment: String
        :param fragment: The fragment string to be added to the last part
        of the URL to be built.
        :type wait: bool
        :param wait: If the engine should wait until the browser URL bar
        is set to the destination URL (could pose issues with HTTP redirection).
        :rtype: bool
        :return: The result of the redirection, if it was verified by
        the browser.
        """

        params_s = []

        for (key, value) in params:
            key_q = appier.util.quote(key)
            for _value in value:
                value_q = appier.util.quote(_value)
                param = key_q + "=" + value_q
                params_s.append(param)

        params_s = "&".join(params_s) if params_s else ""

        if params_s: url += "?" + params_s
        if fragment: url += "#" + fragment

        self.driver.get(url)

        if not wait: return

        return self.waits.redirected_to(redirect_url or url)

    def write_text(self, selector, text):
        """
        Writes the text in the given element, this means having
        it typed like in a physical keyboard.

        :type selector: String
        :param selector: The selector for the element to write the text in.
        :type text: String
        :param text: The text to write "using the keyboard".
        :rtype: Element
        :return: The element with the text changed.
        """

        # waits until the element is visible for the selector and then
        # retrieves the reference to it to be able to write text
        element = self.waits.visible(selector)

        # waits until a valid text change in the element is possible, this
        # overcomes limitations with non interactable elements
        return self.waits.until(
            lambda d: self.driver.safe(self.driver.write_text, element, text),
            "Element '%s' found but never became writable" % selector
        )

    def press_key(self, selector, key):
        """
        Presses the provided key on a certain element, pressed like having
        the proper enter key pressed.

        :type selector: String
        :param selector: The selector for the element to focus when
        pressing enter.
        :type key: String
        :param key: The name of the key that is going to be pressed by
        the keyboard, this name is set on an agnostic way.
        :rtype: Element
        :return: The element with the key pressed.
        """

        # waits until the element is visible for the selector and then
        # retrieves the reference to it to be able to press the key
        element = self.waits.visible(selector)

        # waits until a valid key stroke in the element is possible, this
        # overcomes limitations with non interactable elements
        return self.waits.until(
            lambda d: self.driver.safe(self.driver.press_key, element, key),
            "Element '%s' found but never became interactable" % selector
        )

    def press_enter(self, selector):
        """
        Presses the enter key on a certain element, pressed like having
        the proper enter key pressed.

        :type selector: String
        :param selector: The selector for the element to focus when
        pressing enter.
        :rtype: Element
        :return: The element with the enter key pressed.
        """

        # waits until the element is visible for the selector and then
        # retrieves the reference to it to be able to press enter
        element = self.waits.visible(selector)

        # waits until a valid key stroke in the element is possible, this
        # overcomes limitations with non interactable elements
        return self.waits.until(
            lambda d: self.driver.safe(self.driver.press_enter, element),
            "Element '%s' found but never became interactable" % selector
        )

    def click(self, selector, text = None):
        """
        Clicks an element when possible, which happens when that element is both
        visible and "clickable".

        :type selector: String
        :param selector: The selector for the element to click.
        :type text: String
        :param text: The text the selected element must have before being clicked.
        :rtype: Element
        :return: The clicked element if there's any otherwise an invalid value.
        """

        # waits until the try click operation is possible meaning that a
        # proper click has been "done" by the driver
        return self.waits.until(
            lambda d: self._click(selector, text = text),
            "Element '%s' found but never became clickable" % selector
        )

    def set_file(self, selector, path):
        """
        Sets a file to a file input element when possible, which happens
        when that element exists.

        This operation only sets the file association in the target element
        the upload/send operation should be triggered by the containing
        form element.

        :type selector: String
        :param selector: The selector for the file input element to set
        the file to, the file will only be sent once the form where the
        element is contained is submitted.
        :type path: String
        :param path: The path to the file being sent in the local filesystem.
        :rtype: Element
        :return: The file input element if there's any otherwise an invalid value.
        """

        # normalizes the path (that can be relative) so that it can
        # be safely used by the de underlying element changing operations
        path = os.path.abspath(path)
        path = os.path.normpath(path)

        # waits until the try set operation is possible meaning that
        # the target element exists and the upload was successful
        return self.waits.until(
            lambda d: self._set_file(selector, path),
            "Could not set '%s' to '%s'" % (path, selector)
        )

    def highlight(self, selector, text = None):
        # waits until the element is visible for the selector and then
        # retrieves the reference to it to be able to press enter
        element = self.waits.visible(selector, text = text)

        # waits until the highlight operation is possible for the element
        # that has just been ensured as visible
        return self.waits.until(
            lambda d: self.driver.safe(self.driver.highlight, element),
            "Element '%s' found but was not possible to highlight it" % selector
        )

    def lowlight(self, selector, text = None):
        # waits until the element is visible for the selector and then
        # retrieves the reference to it to be able to press enter
        element = self.waits.visible(selector, text = text)

        # waits until the lowlight operation is possible for the element
        # that has just been ensured as visible
        return self.waits.until(
            lambda d: self.driver.safe(self.driver.lowlight, element),
            "Element '%s' found but was not possible to lowlight it" % selector
        )

    def switch_tab(self, tab):
        return self.driver.switch_tab(tab)

    def switch_context(self, name = "native", index = 0):
        self.waits.until(
            lambda d: self.driver.count_context(name) > index,
            "Expecting the number of contexts to be at least '%d' but is '%d'" % (
                index,
                self.driver.count_context(name)
            )
        )
        return self.driver.switch_context(name, dict(index = index))

    def close_tab(self, tab = None):
        return self.driver.close_tab(tab)

    @property
    def url(self):
        return self.driver.current_url

    def _click(self, selector, text = None):
        """
        Inner method that takes the selector and the possible text value of
        a target element and tries to run a click operation in it.

        This method is ready to be used within a waits environment so that
        proper repetition may happen.

        :type selector: String
        :param selector: The selector for the element to click.
        :type text: String
        :param text: The text the selected element must have before being clicked.
        :rtype: Element
        :return: The clicked element if there's any otherwise an invalid value.
        """

        element = self.waits._get_element(
            selector,
            text = text,
            displayed = False,
            visible = False
        )
        if not element: return None
        return self.driver.safe(self.driver.click, element)

    def _set_file(self, selector, path):
        """
        Inner method that tries to set the file given by path to
        the file input element defined by the selector.

        This method is ready to be used within a waits environment so that
        proper repetition may happen.

        :type selector: String
        :param selector: The selector for the file input element to set
        the file to.
        :type path: String
        :param path: The local filesystem path to the file being sent.
        :rtype: Element
        :return: The file input element if there's any otherwise an invalid value.
        """

        element = self.logic.get(selector)
        if not element: return None
        return self.driver.safe(self.driver.write_text, element, path, False)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys

NAME = "RIPE Rainbow"
VERSION = "0.9.4"
RAINBOW = "🌈"
LABEL = "%s %s %s" % (NAME, VERSION, RAINBOW)
PLATFORM = "%s %d.%d.%d.%s %s" % (
    sys.subversion[0] if hasattr(sys, "subversion") else "CPython",
    sys.version_info[0],
    sys.version_info[1],
    sys.version_info[2],
    sys.version_info[3],
    sys.platform
)
LABEL_FULL = "%s (%s)" % (LABEL, PLATFORM)
USER_AGENTS = dict(
    nexus5 = "Mozilla/5.0 (Linux; Android 4.2.1; en-us; Nexus 5 Build/JOP40D) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.166 Mobile Safari/535.19",
    galaxys3 = "Mozilla/5.0 (Linux; U; Android 4.4.2; en-us; SCH-I535 Build/KOT49H) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30",
    iphone10 = "Mozilla/5.0 (iPhone; CPU iPhone OS 10_3_1 like Mac OS X) AppleWebKit/603.1.30 (KHTML, like Gecko) Version/10.0 Mobile/14E304 Safari/602.1",
    ipad11 = "Mozilla/5.0 (iPad; CPU OS 11_0 like Mac OS X) AppleWebKit/604.1.34 (KHTML, like Gecko) Version/11.0 Mobile/15A5341f Safari/604.1",
    galaxytab3 = "Mozilla/5.0 (Linux; U; Android 4.4.2; de-de; GT-P5210 Build/KOT49H) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Safari/534.30"
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from .. import parts

class WaitsPart(parts.Part):

    def until(self, method, message = None, timeout = None):
        return self.driver.wrap_outer(
            lambda: self.driver._wait(timeout = timeout).until(
                lambda *args: self.driver.wrap_inner(method, *args),
                message = message or "Method was not validated"
            )
        )

    def redirected_to(
        self,
        url,
        params = None,
        fragment = None,
        starts_with = False,
        message = None,
        timeout = None
    ):
        return self.until(
            lambda d: self.logic.at_url(
                url,
                params = params,
                fragment = fragment,
                starts_with = starts_with
            ),
            message = message or "Expecting the page to be '%s' but is '%s'" % (
                url,
                self.driver.current_url
            ),
            timeout = timeout
        )

    def tab_count(self, tab_count, message = None, timeout = None):
        return self.until(
            lambda d: self.driver.tab_count == tab_count,
            message = message or "Expecting the number of browser tabs to be '%s' but is '%s'" % (
                tab_count,
                self.driver.tab_count
            ),
            timeout = timeout
        )

    def has_src(self, selector, url, params = None, ensure = False):
        element = self.waits.visible(selector, ensure = ensure)
        return self.waits.until(
            lambda d: self.logic.has_src(element, url, params = params),
            "'src' attribute was not the expected one"
        )

    def visible(
        self,
        selector,
        text = None,
        ensure = True,
        message = None,
        timeout = None
    ):
        return self.until(
            lambda d: self._get_element(
                selector,
                text = text,
                timeout = timeout,
                displayed = True,
                visible = ensure
            ),
            message = message or "Element '%s' not found or not visible" % selector,
            timeout = timeout
        )

    def not_visible(
        self,
        selector,
        text = None,
        ensure = False,
        message = None,
        timeout = None
    ):
        return self.until(
            lambda d: not self._get_element(
                selector,
                text = text,
                timeout = timeout,
                displayed = True,
                visible = ensure
            ),
            message = message or "Element '%s' was found and is visible" % selector,
            timeout = timeout
        )

    def _get_element(
        self,
        selector,
        text = None,
        timeout = None,
        displayed = True,
        visible = True
    ):
        condition = None
        if text:
            condition = self._build_condition(
                lambda e, s: self.logic.has_text(e, s, text),
                description = "text='%s'" % text
            )
        element = self.logic.get(selector, condition = condition)
        if not element: return None
        if displayed:
            try: self.driver.scroll_to(element)
            except Exception: return None
            result = element.is_displayed()
            if not result: return None
        if visible:
            try: self.driver.ensure_visible(element, timeout = timeout)
            except Exception: return None
        return element

    def _build_condition(self, callable, description = None):
        callable._description = description
        return callable



#!/usr/bin/python
# -*- coding: utf-8 -*-

from . import interactions
from . import logic
from . import waits

from .interactions import InteractionsPart
from .logic import LogicPart
from .waits import WaitsPart



#!/usr/bin/python
# -*- coding: utf-8 -*-

import appier

class SkipError(appier.AppierException):

    def __init__(self, *args, **kwargs):
        appier.AppierException.__init__(self, *args)
        self.reason = kwargs.get("reason", None)

class TimeoutError(appier.AppierException):
    pass

class CloseTabError(appier.AppierException):
    pass

class UnexistingTabError(CloseTabError):

    def __init__(self, *args, **kwargs):
        appier.AppierException.__init__(self, *args)
        tab = kwargs["tab"]
        tab_count = kwargs.get("tab_count", None)
        if tab_count:
            self.reason = "Tab '%s' doesn't exist, there are only %d tabs" %\
                (tab, tab_count)
        else:
            self.reason = "Tab '%s' doesn't exist" % tab



#!/usr/bin/python
# -*- coding: utf-8 -*-

import appier

from .. import parts

class AdminPart(parts.Part):

    def login(self, username, password):
        self.interactions.goto_url(self.signin_url)

        self.interactions.write_text(".form input[name='username']", username)
        self.interactions.write_text(".form input[name='password']", password)
        self.interactions.press_enter(".form input[name='password']")

    def login_wait(self, username, password):
        self.login(username, password)
        self.waits.redirected_to(self.admin_url)

    def click_forgot(self):
        self.interactions.goto_url(self.signin_url)

        self.interactions.click(".forgot a")

    def forgot(self, email):
        self.interactions.goto_url(self.recover_url)

        self.interactions.write_text("input[name='identifier']", email)
        self.interactions.click(".button")

    @property
    def root_url(self):
        ripe_suffix = appier.conf("RIPE_SUFFIX", None)
        if ripe_suffix: root_url = "https://ripe-core-%s.platforme.com" % ripe_suffix
        else: root_url = "http://localhost:8080"
        root_url = appier.conf("BASE_URL", root_url)
        root_url = appier.conf("CORE_URL", root_url)
        root_url = appier.conf("RIPE_CORE_URL", root_url)
        root_url = appier.conf("ADMIN_URL", root_url)
        return appier.conf("RIPE_ADMIN_URL", root_url)

    @property
    def username(self):
        username = appier.conf("CORE_USERNAME", "root")
        username = appier.conf("RIPE_CORE_USERNAME", username)
        username = appier.conf("ADMIN_USERNAME", username)
        username = appier.conf("RIPE_ADMIN_USERNAME", username)
        return username

    @property
    def password(self):
        password = appier.conf("CORE_PASSWORD", "root")
        password = appier.conf("RIPE_CORE_PASSWORD", password)
        password = appier.conf("ADMIN_PASSWORD", password)
        password = appier.conf("RIPE_ADMIN_PASSWORD", password)
        return password

    @property
    def admin_url(self):
        return "%s/admin" % self.root_url

    @property
    def base_url(self):
        return self.admin_url

    @property
    def signin_url(self):
        return "%s/signin" % self.admin_url

    @property
    def recover_url(self):
        return "%s/recover" % self.admin_url
